package java.util.concurrent.locks;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.concurrent.TimeUnit;
import sun.misc.Unsafe;
import java.io.Serializable;

public class StampedLock implements Serializable
{
    private static final long serialVersionUID = -6001602636862214147L;
    private static final int NCPU;
    private static final int SPINS;
    private static final int HEAD_SPINS;
    private static final int MAX_HEAD_SPINS;
    private static final int OVERFLOW_YIELD_RATE = 7;
    private static final int LG_READERS = 7;
    private static final long RUNIT = 1L;
    private static final long WBIT = 128L;
    private static final long RBITS = 127L;
    private static final long RFULL = 126L;
    private static final long ABITS = 255L;
    private static final long SBITS = -128L;
    private static final long ORIGIN = 256L;
    private static final long INTERRUPTED = 1L;
    private static final int WAITING = -1;
    private static final int CANCELLED = 1;
    private static final int RMODE = 0;
    private static final int WMODE = 1;
    private transient volatile StampedLock.WNode whead;
    private transient volatile StampedLock.WNode wtail;
    transient StampedLock.ReadLockView readLockView;
    transient StampedLock.WriteLockView writeLockView;
    transient StampedLock.ReadWriteLockView readWriteLockView;
    private transient volatile long state;
    private transient int readerOverflow;
    private static final Unsafe U;
    private static final long STATE;
    private static final long WHEAD;
    private static final long WTAIL;
    private static final long WNEXT;
    private static final long WSTATUS;
    private static final long WCOWAIT;
    private static final long PARKBLOCKER;
    
    public StampedLock() {
        this.state = 256L;
    }
    
    public long writeLock() {
        long state;
        long n;
        return (((state = this.state) & 0xFFL) == 0x0L && StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n = state + 128L)) ? n : this.acquireWrite(false, 0L);
    }
    
    public long tryWriteLock() {
        long state;
        long n;
        return (((state = this.state) & 0xFFL) == 0x0L && StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n = state + 128L)) ? n : 0L;
    }
    
    public long tryWriteLock(long duration, TimeUnit timeUnit) throws InterruptedException {
        long nanos = timeUnit.toNanos(duration);
        if (!Thread.interrupted()) {
            long tryWriteLock;
            if ((tryWriteLock = this.tryWriteLock()) != 0L) {
                return tryWriteLock;
            }
            if (nanos <= 0L) {
                return 0L;
            }
            long n;
            if ((n = System.nanoTime() + nanos) == 0L) {
                n = 1L;
            }
            long acquireWrite;
            if ((acquireWrite = this.acquireWrite(true, n)) != 1L) {
                return acquireWrite;
            }
        }
        throw new InterruptedException();
    }
    
    public long writeLockInterruptibly() throws InterruptedException {
        long acquireWrite;
        if (!Thread.interrupted() && (acquireWrite = this.acquireWrite(true, 0L)) != 1L) {
            return acquireWrite;
        }
        throw new InterruptedException();
    }
    
    public long readLock() {
        long state = this.state;
        long n;
        return (this.whead == this.wtail && (state & 0xFFL) < 126L && StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n = state + 1L)) ? n : this.acquireRead(false, 0L);
    }
    
    public long tryReadLock() {
        long state;
        long n;
        while ((n = ((state = this.state) & 0xFFL)) != 128L) {
            if (n < 126L) {
                long n2;
                if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n2 = state + 1L)) {
                    return n2;
                }
                continue;
            }
            else {
                long tryIncReaderOverflow;
                if ((tryIncReaderOverflow = this.tryIncReaderOverflow(state)) != 0L) {
                    return tryIncReaderOverflow;
                }
                continue;
            }
        }
        return 0L;
    }
    
    public long tryReadLock(long duration, TimeUnit timeUnit) throws InterruptedException {
        long nanos = timeUnit.toNanos(duration);
        if (!Thread.interrupted()) {
            long state;
            long n;
            if ((n = ((state = this.state) & 0xFFL)) != 128L) {
                if (n < 126L) {
                    long n2;
                    if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n2 = state + 1L)) {
                        return n2;
                    }
                }
                else {
                    long tryIncReaderOverflow;
                    if ((tryIncReaderOverflow = this.tryIncReaderOverflow(state)) != 0L) {
                        return tryIncReaderOverflow;
                    }
                }
            }
            if (nanos <= 0L) {
                return 0L;
            }
            long n3;
            if ((n3 = System.nanoTime() + nanos) == 0L) {
                n3 = 1L;
            }
            long acquireRead;
            if ((acquireRead = this.acquireRead(true, n3)) != 1L) {
                return acquireRead;
            }
        }
        throw new InterruptedException();
    }
    
    public long readLockInterruptibly() throws InterruptedException {
        long acquireRead;
        if (!Thread.interrupted() && (acquireRead = this.acquireRead(true, 0L)) != 1L) {
            return acquireRead;
        }
        throw new InterruptedException();
    }
    
    public long tryOptimisticRead() {
        long state;
        return (((state = this.state) & 0x80L) == 0x0L) ? (state & 0xFFFFFFFFFFFFFF80L) : 0L;
    }
    
    public boolean validate(long n) {
        StampedLock.U.loadFence();
        return (n & 0xFFFFFFFFFFFFFF80L) == (this.state & 0xFFFFFFFFFFFFFF80L);
    }
    
    public void unlockWrite(long n) {
        if (this.state != n || (n & 0x80L) == 0x0L) {
            throw new IllegalMonitorStateException();
        }
        this.state = (((n += 128L) == 0L) ? 256L : n);
        StampedLock.WNode whead;
        if ((whead = this.whead) != null && whead.status != 0) {
            this.release(whead);
        }
        return;
    }
    
    public void unlockRead(long n) {
        long state;
        long n2;
        while (((state = this.state) & 0xFFFFFFFFFFFFFF80L) == (n & 0xFFFFFFFFFFFFFF80L) && (n & 0xFFL) != 0x0L && (n2 = (state & 0xFFL)) != 0L && n2 != 128L) {
            if (n2 < 126L) {
                if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, state - 1L)) {
                    continue;
                }
                StampedLock.WNode whead;
                if (n2 == 1L && (whead = this.whead) != null && whead.status != 0) {
                    this.release(whead);
                }
            }
            else if (this.tryDecReaderOverflow(state) == 0L) {
                continue;
            }
            return;
        }
        throw new IllegalMonitorStateException();
    }
    
    public void unlock(long n) {
        long n2 = n & 0xFFL;
        long state;
        while (((state = this.state) & 0xFFFFFFFFFFFFFF80L) == (n & 0xFFFFFFFFFFFFFF80L)) {
            long n3;
            if ((n3 = (state & 0xFFL)) == 0L) {
                break;
            }
            if (n3 == 128L) {
                if (n2 != n3) {
                    break;
                }
                long n4;
                this.state = (((n4 = state + 128L) == 0L) ? 256L : n4);
                StampedLock.WNode whead2;
                if ((whead2 = this.whead) != null && whead2.status != 0) {
                    this.release(whead2);
                }
                return;
            }
            else if (n2 != 0L) {
                if (n2 >= 128L) {
                    break;
                }
                if (n3 < 126L) {
                    if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, state - 1L)) {
                        continue;
                    }
                    StampedLock.WNode whead;
                    if (n3 == 1L && (whead = this.whead) != null && whead.status != 0) {
                        this.release(whead);
                    }
                    return;
                }
                else {
                    if (this.tryDecReaderOverflow(state) == 0L) {
                        continue;
                    }
                    return;
                }
            }
            else {
                break;
            }
        }
        throw new IllegalMonitorStateException();
    }
    
    public long tryConvertToWriteLock(long n) {
        long n2 = n & 0xFFL;
        long state;
        while (((state = this.state) & 0xFFFFFFFFFFFFFF80L) == (n & 0xFFFFFFFFFFFFFF80L)) {
            long n3;
            if ((n3 = (state & 0xFFL)) == 0L) {
                if (n2 != 0L) {
                    break;
                }
                long n5;
                if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n5 = state + 128L)) {
                    continue;
                }
                return n5;
            }
            else if (n3 == 128L) {
                if (n2 != n3) {
                    break;
                }
                return n;
            }
            else if (n3 == 1L && n2 != 0L) {
                long n4;
                if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n4 = state - 1L + 128L)) {
                    continue;
                }
                return n4;
            }
            else {
                break;
            }
        }
        return 0L;
    }
    
    public long tryConvertToReadLock(long n) {
        long n2 = n & 0xFFL;
        long state;
        while (((state = this.state) & 0xFFFFFFFFFFFFFF80L) == (n & 0xFFFFFFFFFFFFFF80L)) {
            long n3;
            if ((n3 = (state & 0xFFL)) == 0L) {
                if (n2 != 0L) {
                    break;
                }
                if (n3 < 126L) {
                    long n5;
                    if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n5 = state + 1L)) {
                        continue;
                    }
                    return n5;
                }
                else {
                    long tryIncReaderOverflow;
                    if ((tryIncReaderOverflow = this.tryIncReaderOverflow(state)) == 0L) {
                        continue;
                    }
                    return tryIncReaderOverflow;
                }
            }
            else if (n3 == 128L) {
                if (n2 != n3) {
                    break;
                }
                long n4 = this.state = state + 129L;
                StampedLock.WNode whead;
                if ((whead = this.whead) != null && whead.status != 0) {
                    this.release(whead);
                }
                return n4;
            }
            else {
                if (n2 != 0L && n2 < 128L) {
                    return n;
                }
                break;
            }
        }
        return 0L;
    }
    
    public long tryConvertToOptimisticRead(long n) {
        long n2 = n & 0xFFL;
        StampedLock.U.loadFence();
        long state;
        while (((state = this.state) & 0xFFFFFFFFFFFFFF80L) == (n & 0xFFFFFFFFFFFFFF80L)) {
            long n3;
            if ((n3 = (state & 0xFFL)) == 0L) {
                if (n2 == 0L) {
                    return state;
                }
            }
            else if (n3 == 128L) {
                if (n2 == n3) {
                    long n6;
                    long n5 = this.state = (((n6 = state + 128L) == 0L) ? 256L : n6);
                    StampedLock.WNode whead2;
                    if ((whead2 = this.whead) != null && whead2.status != 0) {
                        this.release(whead2);
                    }
                    return n5;
                }
            }
            else if (n2 != 0L && n2 < 128L) {
                if (n3 < 126L) {
                    long n4;
                    if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n4 = state - 1L)) {
                        continue;
                    }
                    StampedLock.WNode whead;
                    if (n3 == 1L && (whead = this.whead) != null && whead.status != 0) {
                        this.release(whead);
                    }
                    return n4 & 0xFFFFFFFFFFFFFF80L;
                }
                else {
                    long tryDecReaderOverflow;
                    if ((tryDecReaderOverflow = this.tryDecReaderOverflow(state)) == 0L) {
                        continue;
                    }
                    return tryDecReaderOverflow & 0xFFFFFFFFFFFFFF80L;
                }
            }
            return 0L;
        }
        return 0L;
    }
    
    public boolean tryUnlockWrite() {
        long state;
        if (((state = this.state) & 0x80L) != 0x0L) {
            long n;
            this.state = (((n = state + 128L) == 0L) ? 256L : n);
            StampedLock.WNode whead;
            if ((whead = this.whead) != null && whead.status != 0) {
                this.release(whead);
            }
            return true;
        }
        else {
            return false;
        }
    }
    
    public boolean tryUnlockRead() {
        long state;
        long n;
        while ((n = ((state = this.state) & 0xFFL)) != 0L && n < 128L) {
            if (n < 126L) {
                if (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, state - 1L)) {
                    continue;
                }
                StampedLock.WNode whead;
                if (n == 1L && (whead = this.whead) != null && whead.status != 0) {
                    this.release(whead);
                }
                return true;
            }
            else {
                if (this.tryDecReaderOverflow(state) == 0L) {
                    continue;
                }
                return true;
            }
        }
        return false;
    }
    
    private int getReadLockCount(long n) {
        long n2;
        if ((n2 = (n & 0x7FL)) >= 126L) {
            n2 = 126L + (long)this.readerOverflow;
        }
        return (int)n2;
    }
    
    public boolean isWriteLocked() {
        return (this.state & 0x80L) != 0x0L;
    }
    
    public boolean isReadLocked() {
        return (this.state & 0x7FL) != 0x0L;
    }
    
    public int getReadLockCount() {
        return this.getReadLockCount(this.state);
    }
    
    @Override
    public String toString() {
        long state = this.state;
        return super.toString() + (((state & 0xFFL) == 0x0L) ? "[Unlocked]" : (((state & 0x80L) != 0x0L) ? "[Write-locked]" : ("[Read-locks:" + this.getReadLockCount(state) + "]")));
    }
    
    public Lock asReadLock() {
        StampedLock.ReadLockView readLockView;
        return ((readLockView = this.readLockView) != null) ? readLockView : (this.readLockView = new StampedLock.ReadLockView());
    }
    
    public Lock asWriteLock() {
        StampedLock.WriteLockView writeLockView;
        return ((writeLockView = this.writeLockView) != null) ? writeLockView : (this.writeLockView = new StampedLock.WriteLockView());
    }
    
    public ReadWriteLock asReadWriteLock() {
        StampedLock.ReadWriteLockView readWriteLockView;
        return ((readWriteLockView = this.readWriteLockView) != null) ? readWriteLockView : (this.readWriteLockView = new StampedLock.ReadWriteLockView());
    }
    
    final void unstampedUnlockWrite() {
        long state;
        if (((state = this.state) & 0x80L) == 0x0L) {
            throw new IllegalMonitorStateException();
        }
        long n;
        this.state = (((n = state + 128L) == 0L) ? 256L : n);
        StampedLock.WNode whead;
        if ((whead = this.whead) != null && whead.status != 0) {
            this.release(whead);
        }
        return;
    }
    
    final void unstampedUnlockRead() {
        long state;
        long n;
        while ((n = ((state = this.state) & 0xFFL)) != 0L && n < 128L) {
            if (n < 126L) {
                if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, state - 1L)) {
                    StampedLock.WNode whead;
                    if (n == 1L && (whead = this.whead) != null && whead.status != 0) {
                        this.release(whead);
                        return;
                    }
                    return;
                }
                else {
                    continue;
                }
            }
            else {
                if (this.tryDecReaderOverflow(state) != 0L) {
                    return;
                }
                continue;
            }
        }
        throw new IllegalMonitorStateException();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.state = 256L;
    }
    
    private long tryIncReaderOverflow(long n) {
        if ((n & 0xFFL) == 0x7EL) {
            if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, n, n | 0x7FL)) {
                ++this.readerOverflow;
                return this.state = n;
            }
        }
        else if ((LockSupport.nextSecondarySeed() & 0x7) == 0x0) {
            Thread.yield();
        }
        return 0L;
    }
    
    private long tryDecReaderOverflow(long n) {
        if ((n & 0xFFL) == 0x7EL) {
            if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, n, n | 0x7FL)) {
                int readerOverflow;
                if ((readerOverflow = this.readerOverflow) > 0) {
                    this.readerOverflow = readerOverflow - 1;
                }
                else {
                    --n;
                }
                return this.state = n;
            }
        }
        else if ((LockSupport.nextSecondarySeed() & 0x7) == 0x0) {
            Thread.yield();
        }
        return 0L;
    }
    
    private void release(StampedLock.WNode o) {
        if (o != null) {
            StampedLock.U.compareAndSwapInt(o, StampedLock.WSTATUS, -1, 0);
            StampedLock.WNode next;
            if ((next = o.next) == null || next.status == 1) {
                for (StampedLock.WNode wNode = this.wtail; wNode != null && wNode != o; wNode = wNode.prev) {
                    if (wNode.status <= 0) {
                        next = wNode;
                    }
                }
            }
            Thread thread;
            if (next != null && (thread = next.thread) != null) {
                StampedLock.U.unpark(thread);
            }
        }
    }
    
    private long acquireWrite(boolean b, long n) {
        Object next = null;
        int n2 = -1;
        while (true) {
            long state;
            long n3;
            if ((n3 = ((state = this.state) & 0xFFL)) == 0L) {
                long n5;
                if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, n5 = state + 128L)) {
                    return n5;
                }
                continue;
            }
            else if (n2 < 0) {
                n2 = ((n3 == 128L && this.wtail == this.whead) ? StampedLock.SPINS : 0);
            }
            else if (n2 > 0) {
                if (LockSupport.nextSecondarySeed() >= 0) {
                    --n2;
                }
                else {
                    continue;
                }
            }
            else if (this.wtail == null) {
                StampedLock.WNode o = new StampedLock.WNode(1);
                if (StampedLock.U.compareAndSwapObject(this, StampedLock.WHEAD, null, o)) {
                    this.wtail = o;
                }
                else {
                    continue;
                }
            }
            else if (next == null) {
                StampedLock.WNode prev;
                next = new StampedLock.WNode(1, prev);
            }
            else {
                StampedLock.WNode prev;
                if (next.prev != prev) {
                    next.prev = prev;
                }
                else if (StampedLock.U.compareAndSwapObject(this, StampedLock.WTAIL, prev, next)) {
                    prev.next = next;
                    int head_SPINS = -1;
                    while (true) {
                        StampedLock.WNode whead;
                        if ((whead = this.whead) == prev) {
                            if (head_SPINS < 0) {
                                head_SPINS = StampedLock.HEAD_SPINS;
                            }
                            else if (head_SPINS < StampedLock.MAX_HEAD_SPINS) {
                                head_SPINS <<= 1;
                            }
                            while (true) {
                                long state2;
                                if (((state2 = this.state) & 0xFFL) == 0x0L) {
                                    long n4;
                                    if (StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state2, n4 = state2 + 128L)) {
                                        this.whead = next;
                                        next.prev = null;
                                        return n4;
                                    }
                                    continue;
                                }
                                else {
                                    if (LockSupport.nextSecondarySeed() >= 0 && --head_SPINS <= 0) {
                                        break;
                                    }
                                    continue;
                                }
                            }
                        }
                        else if (whead != null) {
                            StampedLock.WNode cowait;
                            while ((cowait = whead.cowait) != null) {
                                Thread thread;
                                if (!StampedLock.U.compareAndSwapObject(whead, StampedLock.WCOWAIT, cowait, cowait.cowait) || (thread = cowait.thread) == null) {
                                    continue;
                                }
                                StampedLock.U.unpark(thread);
                            }
                        }
                        if (this.whead == whead) {
                            if ((prev = next.prev) != prev) {
                                if (prev != null) {
                                    prev.next = next;
                                }
                                else {
                                    continue;
                                }
                            }
                            else {
                                int status;
                                if ((status = prev.status) == 0) {
                                    StampedLock.U.compareAndSwapInt(prev, StampedLock.WSTATUS, 0, -1);
                                }
                                else if (status == 1) {
                                    StampedLock.WNode prev2;
                                    if ((prev2 = prev.prev) != null) {
                                        next.prev = prev2;
                                        prev2.next = next;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                                else {
                                    long time;
                                    if (n == 0L) {
                                        time = 0L;
                                    }
                                    else if ((time = n - System.nanoTime()) <= 0L) {
                                        return this.cancelWaiter(next, next, false);
                                    }
                                    Thread currentThread = Thread.currentThread();
                                    StampedLock.U.putObject(currentThread, StampedLock.PARKBLOCKER, this);
                                    next.thread = currentThread;
                                    if (prev.status < 0 && (prev != whead || (this.state & 0xFFL) != 0x0L) && this.whead == whead && next.prev == prev) {
                                        StampedLock.U.park(false, time);
                                    }
                                    next.thread = null;
                                    StampedLock.U.putObject(currentThread, StampedLock.PARKBLOCKER, null);
                                    if (b && Thread.interrupted()) {
                                        return this.cancelWaiter(next, next, true);
                                    }
                                    continue;
                                }
                            }
                        }
                        else {
                            continue;
                        }
                    }
                }
                else {
                    continue;
                }
            }
        }
    }
    
    private long acquireRead(boolean b, long n) {
        Object o = null;
        int spins = -1;
        long tryIncReaderOverflow;
    Label_0092:
        while (true) {
            StampedLock.WNode whead;
            StampedLock.WNode wtail;
            if (this.whead == this.wtail) {
                while (true) {
                    long state;
                    long n2;
                    if (!(((n2 = ((state = this.state) & 0xFFL)) >= 126L) ? (n2 >= 128L || (tryIncReaderOverflow = this.tryIncReaderOverflow(state)) == 0L) : (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state, tryIncReaderOverflow = state + 1L)))) {
                        break Label_0092;
                    }
                    if (n2 < 128L) {
                        continue;
                    }
                    if (spins > 0) {
                        if (LockSupport.nextSecondarySeed() < 0) {
                            continue;
                        }
                        --spins;
                    }
                    else {
                        if (spins == 0) {
                            whead = this.whead;
                            wtail = this.wtail;
                            if (whead != whead || wtail != wtail) {
                                if (whead != wtail) {
                                    break;
                                }
                            }
                            else {
                                break;
                            }
                        }
                        spins = StampedLock.SPINS;
                    }
                }
            }
            if (wtail == null) {
                StampedLock.WNode o3 = new StampedLock.WNode(1);
                if (StampedLock.U.compareAndSwapObject(this, StampedLock.WHEAD, null, o3)) {
                    this.wtail = o3;
                }
                else {
                    continue;
                }
            }
            else if (o == null) {
                o = new StampedLock.WNode(0, wtail);
            }
            else if (whead == wtail || wtail.mode != 0) {
                if (o.prev != wtail) {
                    o.prev = wtail;
                }
                else {
                    if (StampedLock.U.compareAndSwapObject(this, StampedLock.WTAIL, wtail, o)) {
                        wtail.next = o;
                        int head_SPINS = -1;
                        long tryIncReaderOverflow3;
                    Label_0815:
                        while (true) {
                            StampedLock.WNode whead3;
                            if ((whead3 = this.whead) == wtail) {
                                if (head_SPINS < 0) {
                                    head_SPINS = StampedLock.HEAD_SPINS;
                                }
                                else if (head_SPINS < StampedLock.MAX_HEAD_SPINS) {
                                    head_SPINS <<= 1;
                                }
                                while (true) {
                                    long state3;
                                    long n4;
                                    if (!(((n4 = ((state3 = this.state) & 0xFFL)) >= 126L) ? (n4 >= 128L || (tryIncReaderOverflow3 = this.tryIncReaderOverflow(state3)) == 0L) : (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state3, tryIncReaderOverflow3 = state3 + 1L)))) {
                                        break Label_0815;
                                    }
                                    if (n4 >= 128L && LockSupport.nextSecondarySeed() >= 0 && --head_SPINS <= 0) {
                                        break;
                                    }
                                    continue;
                                }
                            }
                            else if (whead3 != null) {
                                StampedLock.WNode cowait3;
                                while ((cowait3 = whead3.cowait) != null) {
                                    Thread thread2;
                                    if (!StampedLock.U.compareAndSwapObject(whead3, StampedLock.WCOWAIT, cowait3, cowait3.cowait) || (thread2 = cowait3.thread) == null) {
                                        continue;
                                    }
                                    StampedLock.U.unpark(thread2);
                                }
                            }
                            if (this.whead == whead3) {
                                StampedLock.WNode prev2;
                                if ((prev2 = o.prev) != wtail) {
                                    if (prev2 != null) {
                                        prev2.next = o;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                                else {
                                    int status;
                                    if ((status = wtail.status) == 0) {
                                        StampedLock.U.compareAndSwapInt(wtail, StampedLock.WSTATUS, 0, -1);
                                    }
                                    else if (status == 1) {
                                        StampedLock.WNode prev3;
                                        if ((prev3 = wtail.prev) != null) {
                                            o.prev = prev3;
                                            prev3.next = o;
                                        }
                                        else {
                                            continue;
                                        }
                                    }
                                    else {
                                        long time2;
                                        if (n == 0L) {
                                            time2 = 0L;
                                        }
                                        else if ((time2 = n - System.nanoTime()) <= 0L) {
                                            return this.cancelWaiter(o, o, false);
                                        }
                                        Thread currentThread2 = Thread.currentThread();
                                        StampedLock.U.putObject(currentThread2, StampedLock.PARKBLOCKER, this);
                                        o.thread = currentThread2;
                                        if (wtail.status < 0 && (wtail != whead3 || (this.state & 0xFFL) == 0x80L) && this.whead == whead3 && o.prev == wtail) {
                                            StampedLock.U.park(false, time2);
                                        }
                                        o.thread = null;
                                        StampedLock.U.putObject(currentThread2, StampedLock.PARKBLOCKER, null);
                                        if (b && Thread.interrupted()) {
                                            return this.cancelWaiter(o, o, true);
                                        }
                                        continue;
                                    }
                                }
                            }
                            else {
                                continue;
                            }
                        }
                        this.whead = o;
                        o.prev = null;
                        StampedLock.WNode cowait4;
                        while ((cowait4 = o.cowait) != null) {
                            Thread thread3;
                            if (!StampedLock.U.compareAndSwapObject(o, StampedLock.WCOWAIT, cowait4, cowait4.cowait) || (thread3 = cowait4.thread) == null) {
                                continue;
                            }
                            StampedLock.U.unpark(thread3);
                        }
                        return tryIncReaderOverflow3;
                    }
                    continue;
                }
            }
            else {
                Unsafe u = StampedLock.U;
                Object o2 = wtail;
                long wcowait = StampedLock.WCOWAIT;
                StampedLock.WNode cowait = wtail.cowait;
                o.cowait = cowait;
                if (!u.compareAndSwapObject(o2, wcowait, cowait, o)) {
                    o.cowait = null;
                }
                else {
                    while (true) {
                        StampedLock.WNode whead2;
                        StampedLock.WNode cowait2;
                        Thread thread;
                        if ((whead2 = this.whead) != null && (cowait2 = whead2.cowait) != null && StampedLock.U.compareAndSwapObject(whead2, StampedLock.WCOWAIT, cowait2, cowait2.cowait) && (thread = cowait2.thread) != null) {
                            StampedLock.U.unpark(thread);
                        }
                        StampedLock.WNode prev;
                        if (whead2 == (prev = wtail.prev) || whead2 == wtail || prev == null) {
                            long n3;
                            do {
                                long state2;
                                long tryIncReaderOverflow2;
                                if (!(((n3 = ((state2 = this.state) & 0xFFL)) >= 126L) ? (n3 >= 128L || (tryIncReaderOverflow2 = this.tryIncReaderOverflow(state2)) == 0L) : (!StampedLock.U.compareAndSwapLong(this, StampedLock.STATE, state2, tryIncReaderOverflow2 = state2 + 1L)))) {
                                    return tryIncReaderOverflow2;
                                }
                            } while (n3 < 128L);
                        }
                        if (this.whead == whead2 && wtail.prev == prev) {
                            if (prev == null || whead2 == wtail || wtail.status > 0) {
                                break;
                            }
                            long time;
                            if (n == 0L) {
                                time = 0L;
                            }
                            else if ((time = n - System.nanoTime()) <= 0L) {
                                return this.cancelWaiter(o, wtail, false);
                            }
                            Thread currentThread = Thread.currentThread();
                            StampedLock.U.putObject(currentThread, StampedLock.PARKBLOCKER, this);
                            o.thread = currentThread;
                            if ((whead2 != prev || (this.state & 0xFFL) == 0x80L) && this.whead == whead2 && wtail.prev == prev) {
                                StampedLock.U.park(false, time);
                            }
                            o.thread = null;
                            StampedLock.U.putObject(currentThread, StampedLock.PARKBLOCKER, null);
                            if (b && Thread.interrupted()) {
                                return this.cancelWaiter(o, wtail, true);
                            }
                            continue;
                        }
                        else {
                            continue;
                        }
                    }
                    o = null;
                    continue;
                }
            }
        }
        return tryIncReaderOverflow;
    }
    
    private long cancelWaiter(StampedLock.WNode expected, StampedLock.WNode o, boolean b) {
        if (expected != null && o != null) {
            expected.status = 1;
            StampedLock.WNode cowait;
            while ((cowait = o.cowait) != null) {
                if (cowait.status == 1) {
                    StampedLock.U.compareAndSwapObject(o, StampedLock.WCOWAIT, cowait, cowait.cowait);
                }
                else {
                    o = cowait;
                }
            }
            if (o == expected) {
                for (StampedLock.WNode wNode = o.cowait; wNode != null; wNode = wNode.cowait) {
                    Thread thread;
                    if ((thread = wNode.thread) != null) {
                        StampedLock.U.unpark(thread);
                    }
                }
                StampedLock.WNode prev = expected.prev;
                while (prev != null) {
                    StampedLock.WNode wNode2;
                    while (expected.next == null || wNode2.status == 1) {
                        wNode2 = null;
                        for (StampedLock.WNode wNode3 = this.wtail; wNode3 != null && wNode3 != expected; wNode3 = wNode3.prev) {
                            if (wNode3.status != 1) {
                                wNode2 = wNode3;
                            }
                        }
                        if (wNode2 == wNode2 || StampedLock.U.compareAndSwapObject(expected, StampedLock.WNEXT, wNode2, wNode2)) {
                            if (wNode2 == null && expected == this.wtail) {
                                StampedLock.U.compareAndSwapObject(this, StampedLock.WTAIL, expected, prev);
                                break;
                            }
                            break;
                        }
                        else {
                            continue;
                        }
                    }
                    if (prev.next == expected) {
                        StampedLock.U.compareAndSwapObject(prev, StampedLock.WNEXT, expected, wNode2);
                    }
                    Thread thread2;
                    if (wNode2 != null && (thread2 = wNode2.thread) != null) {
                        wNode2.thread = null;
                        StampedLock.U.unpark(thread2);
                    }
                    if (prev.status == 1) {
                        StampedLock.WNode prev2;
                        if ((prev2 = prev.prev) == null) {
                            break;
                        }
                        expected.prev = prev2;
                        StampedLock.U.compareAndSwapObject(prev2, StampedLock.WNEXT, prev, wNode2);
                        prev = prev2;
                    }
                    else {
                        break;
                    }
                }
            }
        }
        StampedLock.WNode whead;
        while ((whead = this.whead) != null) {
            StampedLock.WNode next;
            if ((next = whead.next) == null || next.status == 1) {
                for (StampedLock.WNode wNode4 = this.wtail; wNode4 != null && wNode4 != whead; wNode4 = wNode4.prev) {
                    if (wNode4.status <= 0) {
                        next = wNode4;
                    }
                }
            }
            if (whead == this.whead) {
                long state;
                if (next != null && whead.status == 0 && ((state = this.state) & 0xFFL) != 0x80L && (state == 0L || next.mode == 0)) {
                    this.release(whead);
                    break;
                }
                break;
            }
            else {
                continue;
            }
        }
        return b || Thread.interrupted()?1L:0L;
    }
    
    static {
        NCPU = Runtime.getRuntime().availableProcessors();
        SPINS = ((StampedLock.NCPU > 1) ? 64 : 0);
        HEAD_SPINS = ((StampedLock.NCPU > 1) ? 1024 : 0);
        MAX_HEAD_SPINS = ((StampedLock.NCPU > 1) ? 65536 : 0);
        try {
            U = Unsafe.getUnsafe();
            Class<Object> clazz = StampedLock.class;
            Class<Object> clazz2 = StampedLock.WNode.class;
            STATE = StampedLock.U.objectFieldOffset(clazz.getDeclaredField("state"));
            WHEAD = StampedLock.U.objectFieldOffset(clazz.getDeclaredField("whead"));
            WTAIL = StampedLock.U.objectFieldOffset(clazz.getDeclaredField("wtail"));
            WSTATUS = StampedLock.U.objectFieldOffset(clazz2.getDeclaredField("status"));
            WNEXT = StampedLock.U.objectFieldOffset(clazz2.getDeclaredField("next"));
            WCOWAIT = StampedLock.U.objectFieldOffset(clazz2.getDeclaredField("cowait"));
            PARKBLOCKER = StampedLock.U.objectFieldOffset(Thread.class.getDeclaredField("parkBlocker"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class WNode
    {
        volatile StampedLock.WNode prev;
        volatile StampedLock.WNode next;
        volatile StampedLock.WNode cowait;
        volatile Thread thread;
        volatile int status;
        final int mode;
        
        WNode(int mode, StampedLock.WNode prev) {
            this.mode = mode;
            this.prev = prev;
        }
    }
    
    final class ReadLockView implements Lock
    {
        @Override
        public void lock() {
            StampedLock.this.readLock();
        }
        
        @Override
        public void lockInterruptibly() throws InterruptedException {
            StampedLock.this.readLockInterruptibly();
        }
        
        @Override
        public boolean tryLock() {
            return StampedLock.this.tryReadLock() != 0L;
        }
        
        @Override
        public boolean tryLock(long n, TimeUnit timeUnit) throws InterruptedException {
            return StampedLock.this.tryReadLock(n, timeUnit) != 0L;
        }
        
        @Override
        public void unlock() {
            StampedLock.this.unstampedUnlockRead();
        }
        
        @Override
        public Condition newCondition() {
            throw new UnsupportedOperationException();
        }
    }
    
    final class WriteLockView implements Lock
    {
        @Override
        public void lock() {
            StampedLock.this.writeLock();
        }
        
        @Override
        public void lockInterruptibly() throws InterruptedException {
            StampedLock.this.writeLockInterruptibly();
        }
        
        @Override
        public boolean tryLock() {
            return StampedLock.this.tryWriteLock() != 0L;
        }
        
        @Override
        public boolean tryLock(long n, TimeUnit timeUnit) throws InterruptedException {
            return StampedLock.this.tryWriteLock(n, timeUnit) != 0L;
        }
        
        @Override
        public void unlock() {
            StampedLock.this.unstampedUnlockWrite();
        }
        
        @Override
        public Condition newCondition() {
            throw new UnsupportedOperationException();
        }
    }
    
    final class ReadWriteLockView implements ReadWriteLock
    {
        @Override
        public Lock readLock() {
            return StampedLock.this.asReadLock();
        }
        
        @Override
        public Lock writeLock() {
            return StampedLock.this.asWriteLock();
        }
    }
}
